@@ -13,7 +13,28 @@ import re
 __all__ = (
     'PythonCompleter',
 )
+non_completable_keywords={'class','finally','is','return','continue','for','lambda','try','def','from','nonlocal','while','and','del','global','not','with','as','elif','if','or','yield','assert','else','import','pass','break','except','in','raise'}
 
+import sys
+_all_module_names=set(sys.builtin_module_names)
+def get_all_importable_module_names():
+    #Returns a set of all names that you can use 'import <name>' on
+    import pkgutil
+    if _all_module_names:
+        def update_thread():
+            #Every time we call this fuction, make a new thread that will update the _all_module_names for us (without blocking)
+            #That way we can respond to new packages being added, or changes in directory etc without having to block completions for them
+            for _,name,_ in pkgutil.iter_modules():
+                _all_module_names.add(name)
+        from threading import Thread
+        Thread(target=update_thread).start()
+        return _all_module_names
+    for _,name,_ in pkgutil.iter_modules():
+        _all_module_names.add(name)
+    return _all_module_names
+
+from threading import Thread
+Thread(target=get_all_importable_module_names).start()
 
 class PythonCompleter(Completer):
     """
@@ -110,7 +131,121 @@ class PythonCompleter(Completer):
         force=True
         global old_origin,candidates
         origin=document.get_word_before_cursor()
+
+        def yield_from_candidates(candidates:list):
+            for c in candidates:
+                yield Completion(text=c,start_position=-len(origin),display=c);
         import rp.r_iterm_comm as ric
+        before_line=document.current_line_before_cursor
+
+        import re
+        from rp import ring_terminal_bell
+        # ring_terminal_bell()
+        if re.fullmatch(r' *(from|import) +\w*',before_line):
+            #Speed up 'import ___|' completions. Note that this is going to skip modules that have been pip_installed during runtime, but this is a small price to pay for this kind of crazy fast autocompletion speed.
+            #This kinda completion is particularly useful because the spacebar triggers the autocpmpletion of module names like tab does
+            yield from yield_from_candidates(get_all_importable_module_names())
+            return
+        # if re.fullmatch(r'.*[^\w\.]',before_line):
+        #     yield from yield_from_candidates(list(ric.globa))
+        #     # from rp import play_tone
+        #     # play_tone(880,seconds=.03)  
+        #     return
+        if not '\n' in document.text and re.fullmatch(r' *((if|with|while|assert|yield|yield from|elif) )? *\w*',before_line):
+            # ring_terminal_bell()
+            yield from yield_from_candidates(list(ric.globa))
+            return
+        blr=before_line[::-1]#blr=before_line reversed
+        if '.' in blr:#This is a hack to make completions to namespaces like np.* not need to use jedi (which takes about 1 second to make a script which is super slow). This is a hack but it makes completions that would otherwise have to use jedi over 20x faster.
+            found=re.match(r'(\.\w*[a-zA-Z_])+',blr[blr.find('.'):])
+            if not found:
+                pass#Something went wrong whatever
+            elif found.string:
+                found=found.string
+                assert found,'wat? found='+repr(found.match)
+                prefix=found[1:][::-1]
+                try:
+                    def keys(root):
+                        out=set()
+                        try:out.update(dir(root))
+                        except:pass
+                        try:out.update(root.__dict__)
+                        except:pass
+                        return sorted(out)
+                    def evall():
+                        parts=prefix.split('.')
+                        cursor=ric.globa[parts[0]]
+                        for part in parts[1:]:
+                            cursor=getattr(cursor,part)
+                        return cursor
+                    yield from yield_from_candidates(keys(evall()))
+                    # yield from yield_from_candidates(keys(eval(prefix,ric.globa)))
+                    # from rp import play_tone
+                    # play_tone(seconds=.03)
+                    # print("HI"+repr(prefix)+"BYE")
+                    return
+                except Exception as E:
+                    # from rp import print_stack_trace
+                    # print_stack_trace(E)
+                    #This type of error is perfectly normal and is to be expected.
+                    #  ⮤ adsfpor.print_stack_trace: ERROR: Traceback (most recent call last):
+                    #   File "/Library/Frameworks/Python.framework/Versions/3.5/lib/python3.5/site-packages/rp/rp_ptpython/completer.py", line 136, in _get_completions                                                                                     
+                    #     yield from yield_from_candidates(keys(eval(prefix,ric.globa)))                                  
+                    #   File "<string>", line 1, in <module>                                                              
+                    # NameError: name 'adsfpo' is not defined  
+                    pass#Oh well whatevs we'll do it with jedi, it will just be a bit slower
+
+        import re
+        current_line=document.current_line
+        before_line=document.current_line_before_cursor
+        after_line=document.current_line_after_cursor
+        after=document.text_after_cursor
+        before=document.text_before_cursor
+        in_global_or_nonlocal_declaration=(re.fullmatch(r'\s*(del|global|nonlocal)\s+[\w\s\,]*',before_line))
+        # if in_global_or_nonlocal_declaration:
+        #     print("GOOGOO")
+            #Make @memoized complete faster:
+            #If we set the mode to 'fast' instead of 'good':
+            #after 'if |' or 'while |' etc, you might want to use fast autocompletions instead of good autocompletions, because I can't think of a situation where it would miss any completions (though it might rip some out of strings and comments). This gives a speed boost.
+        # if re.fullmatch(r'\s*def\s+\w+\s*\((\s*\w+\s*(\=\s*[^\s].*)?\s*\,\s*)*\w*',before_line) and after_line.strip()=='):':
+        #     return#Don't autocomplete function paramater names
+
+        if before_line.startswith('CD ') and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
+            import os
+            from rp import is_a_directory
+            yield from yield_from_candidates([x for x in os.listdir() if is_a_directory(x)])
+            return 
+        if re.fullmatch(r'(CAT |NCAT |CCAT |OPEN ).*',before) and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
+            import os
+            from rp import is_a_file
+            yield from yield_from_candidates([x for x in os.listdir() if is_a_file(x)])
+            return 
+        if re.fullmatch(r'.*\`\w*',before_line)\
+            or (before_line.startswith('RUN ') and not ('\n' in before) and not after):#not after and not '\n' in before and re.fullmatch(before_line):
+            import os
+            yield from yield_from_candidates([x for x in os.listdir() if x.endswith('.py')])
+            return 
+        if (before_line.startswith('!') or before_line.startswith('ARG ')) and not ('\n' in before) and not after:#not after and not '\n' in before and re.fullmatch(before_line):
+            import os
+            yield from yield_from_candidates([x for x in os.listdir()])
+            return 
+
+        import re
+
+        # if re.fullmatch(r'(.*\W)?',before_line)
+
+
+        if  in_global_or_nonlocal_declaration     or\
+            (not after and re.fullmatch(r'\@\w*',before_line)) or\
+            (re.fullmatch(r'\s*(if|while|(for\s+(\w\s\,)+\s+in)|assert|with|return|yield|(yield from)|elif|)\s+\w*',before_line)) or \
+            (ric.completion_style and ric.completion_style[0]=='fast'): #fast but bad completion mode
+            tokens=re.findall(r'[a-zA-Z_]\w*',document.text)#all possible tokens in the current buffer
+            yield from yield_from_candidates((set(tokens)|set(ric.globa))-non_completable_keywords)
+            return
+
+
+        ##################### JEDI SECTION. THIS IS A SLOW AREA  ##############
+
         # ric.current_candidates.clear()
         # if not origin:
         #     return
@@ -185,14 +320,23 @@ class PythonCompleter(Completer):
         # print()
         # print(pre_origin_doc)
         # print()
+        # document=buffer.document
+
+        before_line=document.current_line_before_cursor
+        if not before_line.strip() and not force:
+           return
+
         flag=False
         while True:
             if pre_origin_doc not in completion_cache_pre_origin_doc or flag:
                 ric.current_candidates=[]
+                from rp import tic,ptoctic
+                # tic()
                 for c in self._get_completions(document, complete_event,force=False):
                     # print(c.text)
                     # print(c.text)
                     ric.current_candidates.append(str(c.text))
+                # ptoctic()
                 completion_cache_pre_origin_doc[pre_origin_doc]=tuple(ric.current_candidates)
                 break
             else:
@@ -215,7 +359,11 @@ class PythonCompleter(Completer):
             # ric.current_candidates==ric.current_candidates[::-1]
 
         for x in ric.current_candidates:
-            yield Completion(text=x,start_position=-len(post_period_origin),display=x)
+            if not x.startswith('_'):#Make sure things starting with '_' come last...
+                yield Completion(text=x,start_position=-len(post_period_origin),display=x)
+        for x in ric.current_candidates:
+            if     x.startswith('_'):
+                yield Completion(text=x,start_position=-len(post_period_origin),display=x)
         # print("TIME2:",end='')
         # ptoc()
         # print()
@@ -232,7 +380,6 @@ from functools import lru_cache
 _ryan_completion_matches_cache={}#Note: over time this might cause a memory leakage....idk if that will be a problem....
 def ryan_completion_matches(origin:str,candidates:list):
     import rp
-    rp.tic()
     #This function accomplishes one of the few **beautiful** things in rp.pseudo_terminal: my string matching algorithm, specifically made for python namespace elements
     # Better autocompletions
     # candidates origin is string
